--[[

====================================================
ChronoBars - buff, debuff, cooldown, totem tracker
Author: Ivan Leben
====================================================

--]]

ChronoBars = {}
local CB = ChronoBars;

ChronoBars.VERSION = "1.7";
ChronoBars.UPGRADE_LIST = { "1.2","1.3","1.4","1.5","1.6" };

-- Define constants
--==============================================================

ChronoBars.MSG_PREFIX = "<ChronoBars>";

ChronoBars.EFFECT_TYPE_AURA         = 1;
ChronoBars.EFFECT_TYPE_CD           = 2;
ChronoBars.EFFECT_TYPE_INTERNAL_CD  = 3;
ChronoBars.EFFECT_TYPE_TOTEM        = 4;
ChronoBars.EFFECT_TYPE_USABLE       = 5;
ChronoBars.EFFECT_TYPE_CUSTOM       = 6;

ChronoBars.AURA_UNIT_PLAYER          = "player";
ChronoBars.AURA_UNIT_TARGET          = "target";
ChronoBars.AURA_UNIT_TARGET_TARGET   = "targettarget";
ChronoBars.AURA_UNIT_FOCUS           = "focus";
ChronoBars.AURA_UNIT_FOCUS_TARGET    = "focustarget";
ChronoBars.AURA_UNIT_PET             = "pet";
ChronoBars.AURA_UNIT_PET_TARGET      = "pettarget";

ChronoBars.AURA_TYPE_BUFF    = "HELPFUL";
ChronoBars.AURA_TYPE_DEBUFF  = "HARMFUL";

ChronoBars.CD_TYPE_SPELL      = 1;
ChronoBars.CD_TYPE_PET_SPELL  = 2;
ChronoBars.CD_TYPE_ITEM       = 3;

ChronoBars.USABLE_TYPE_SPELL      = 1;
ChronoBars.USABLE_TYPE_PET_SPELL  = 2;
ChronoBars.USABLE_TYPE_ITEM       = 3;

ChronoBars.TOTEM_TYPE_FIRE    = 1;
ChronoBars.TOTEM_TYPE_EARTH   = 2;
ChronoBars.TOTEM_TYPE_WATER   = 3;
ChronoBars.TOTEM_TYPE_AIR     = 4;

ChronoBars.CUSTOM_TRIGGER_SPELL_CAST = 1;
ChronoBars.CUSTOM_TRIGGER_BAR_ACTIVE = 2;

ChronoBars.SIDE_LEFT    = 1;
ChronoBars.SIDE_RIGHT   = 2;

ChronoBars.TIME_SINGLE   = 1;
ChronoBars.TIME_DECIMAL  = 2;
ChronoBars.TIME_MINSEC   = 3;

ChronoBars.LAYOUT_KEEP      = 1;
ChronoBars.LAYOUT_STACK     = 2;
ChronoBars.LAYOUT_PRIORITY  = 3;

ChronoBars.GROW_UP     = 1;
ChronoBars.GROW_DOWN   = 2;

ChronoBars.VISIBLE_ALWAYS   = 1;
ChronoBars.VISIBLE_ACTIVE   = 2;

ChronoBars.SORT_NONE        = 1;
ChronoBars.SORT_ASCENDING   = 2;
ChronoBars.SORT_DESCENDING  = 3;


-- Define defaults
--=========================================================

ChronoBars.DEFAULT_BAR = {
  name        = "Effect",
  type        = ChronoBars.EFFECT_TYPE_AURA,
  aura        = {
                  unit = ChronoBars.AURA_UNIT_PLAYER,
                  type = ChronoBars.AURA_TYPE_BUFF,
                  byPlayer = false,
                  sum = false,
                  order = 1,
                },
  cd          = {
                  type = ChronoBars.CD_TYPE_SPELL,
                },
  usable      = {
                  type = ChronoBars.USABLE_TYPE_SPELL,
                  includeCd = true,
                },
  totem       = {
                  type = ChronoBars.TOTEM_TYPE_FIRE,
                },
  custom      = {
                  trigger = ChronoBars.CUSTOM_TRIGGER_SPELL_CAST,
                  duration = 0,
                },
  display     = {
                  enabled = false,
                  name = "DisplayName",
                },
  fixed       = {
                  enabled = false,
                  duration = 0,
                },
  style       = {
                  showName       = true,
                  showIcon       = true,
                  showTime       = true,
                  showCount      = true,
                  showCd         = true,
                  showUsable     = true,
                  showSpark      = true,
                  timeSide       = ChronoBars.SIDE_RIGHT,
                  countSide      = ChronoBars.SIDE_RIGHT,
                  fullSide       = ChronoBars.SIDE_RIGHT,
                  iconSide       = ChronoBars.SIDE_LEFT,
                  iconZoom       = true,
                  fillUp         = false,
                  fgColor        = { r = 0.0, g = 0.4, b = 0.8, a = 1.0 },
                  bgColor        = { r = 0.0, g = 0.0, b = 0.0, a = 0.7 },
                  textColor      = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
                  fontSize       = 12,
                  sparkHeight    = 1.8,
                  sparkWidth     = 20,
                  timeFormat     = ChronoBars.TIME_SINGLE,
                  lsmFontHandle  = "Friz Quadrata TT",
                  lsmTexHandle   = "None",
                  visibility  = ChronoBars.VISIBLE_ACTIVE,
                  anim  = {
                            up = true,
                            down = true,
                            blink = true,
                            blinkUsable = true,
                            fade = true,
                          },
                },
};

ChronoBars.DEFAULT_GROUP = {
  bars         = {},
  x            = 0,
  y            = 0,
  width        = 200,
  height       = 20,
  padding      = 0,
  spacing      = 0,
  margin       = 5,
  grow         = ChronoBars.GROW_UP,
  layout       = ChronoBars.LAYOUT_STACK,
  sorting      = ChronoBars.SORT_NONE,
  style        = {
                   bgColor      = { r = 0.0, g = 0.0, b = 0.0, a = 0.0 },
                 },
};

ChronoBars.DEFAULT_PROFILE = {
  groups       = {},
};

ChronoBars.DEFAULT_SETTINGS = {
  version         = ChronoBars.VERSION,
  updateInterval  = 0.03,
  debugEnabled    = false,
  profiles        = {},
}

ChronoBars.DEFAULT_CHAR_SETTINGS = {
  version          = ChronoBars.VERSION,
  activeProfile    = "Default",
  primaryProfile   = nil,
  secondaryProfile = nil,
}


table.insert( ChronoBars.DEFAULT_GROUP.bars, ChronoBars.DEFAULT_BAR );
table.insert( ChronoBars.DEFAULT_PROFILE.groups, ChronoBars.DEFAULT_GROUP );
ChronoBars.DEFAULT_SETTINGS.profiles[ "Default" ] = ChronoBars.DEFAULT_PROFILE;


--Upgrade functions
--=====================================================

-- 1.2

function ChronoBars.Upgrade_1_2 ()

  if (ChronoBars_Settings.updateInterval == nil) then
    ChronoBars_Settings.updateInterval = 0.03;
  end
  
  for pname, profile in pairs( ChronoBars_Settings.profiles ) do
    for g=1,table.getn( profile.groups ) do
      for b=1,table.getn( profile.groups[g].bars ) do
      
        local bar = profile.groups[g].bars[b];
        
        if (bar.style.showCd == nil) then
          bar.style.showCd = bar.style.appendCd;
        end
        
        if (bar.style.showCount == nil) then
          bar.style.showCount = true;
        end
        
        if (bar.style.countSide == nil) then
          bar.style.countSide = ChronoBars.SIDE_RIGHT;
        end

        if (bar.style.lsmFontHandle == nil) then
          bar.style.lsmFontHandle  = "Friz Quadrata TT";
        end

        if (bar.style.lsmTexHandle == nil) then
          bar.style.lsmTexHandle   = "None";
        end
        
        if (bar.style.iconZoom == nil) then
          bar.style.iconZoom = true;
        end
        
      end
    end
  end
end

function ChronoBars.UpgradeChar_1_2 ()
end

-- 1.3

function ChronoBars.Upgrade_1_3 ()

  for pname, profile in pairs( ChronoBars_Settings.profiles ) do
    for g=1,table.getn( profile.groups ) do
      local grp = profile.groups[g];
      
      if (grp.style == nil) then
        grp.style = CopyTable( ChronoBars.DEFAULT_GROUP.style );
      end
      
      if (grp.margin == nil) then
        grp.margin = ChronoBars.DEFAULT_GROUP.margin;
      end

      for b=1,table.getn( profile.groups[g].bars ) do
        local bar = profile.groups[g].bars[b];
        
      end
    end
  end
end

function ChronoBars.UpgradeChar_1_3 ()
end


-- 1.4

function ChronoBars.Upgrade_1_4 ()

  for pname, profile in pairs( ChronoBars_Settings.profiles ) do
    for g=1,table.getn( profile.groups ) do
      local grp = profile.groups[g];

      for b=1,table.getn( profile.groups[g].bars ) do
        local bar = profile.groups[g].bars[b];

        if (bar.aura.order == nil) then
          bar.aura.order = 1;
        end
        
        if (bar.usable == nil) then
          bar.usable = CopyTable( CB.DEFAULT_BAR.usable );
        end
        
        if (bar.style.showUsable == nil) then
          bar.style.showUsable = true;
        end
        
      end
    end
  end
end

function ChronoBars.UpgradeChar_1_4 ()
end


-- 1.5

function ChronoBars.Upgrade_1_5 ()

  if (ChronoBars_Settings.updateInterval == 0.03) then
    ChronoBars_Settings.updateInterval = 0.01;
  end
  
  for pname, profile in pairs( ChronoBars_Settings.profiles ) do
    for g=1,table.getn( profile.groups ) do
      local grp = profile.groups[g];

      for b=1,table.getn( profile.groups[g].bars ) do
        local bar = profile.groups[g].bars[b];
        
        if (bar.display == nil) then
          bar.display = CopyTable( CB.DEFAULT_BAR.display );
        end
        
        if (bar.custom == nil) then
          bar.custom = CopyTable( CB.DEFAULT_BAR.custom );
        end
        
        if (bar.type == CB.EFFECT_TYPE_INTERNAL_CD) then
          bar.type = CB.EFFECT_TYPE_CUSTOM;
          bar.custom.trigger = CB.CUSTOM_TRIGGER_BAR_ACTIVE;
          bar.custom.duration = bar.internal.duration;
        end
        
        if (bar.style.showSpark == nil) then
          bar.style.showSpark = true;
        end
        
        if (bar.style.anim == nil) then
          bar.style.anim = CopyTable( CB.DEFAULT_BAR.style.anim );
        end
        
        if (bar.style.anim.blinkUsable == nil) then
          bar.style.anim.blinkUsable = true;
        end
        
      end
    end
  end
end

function ChronoBars.UpgradeChar_1_5 ()
end

-- 1.6

function ChronoBars.Upgrade_1_6 ()

  for pname, profile in pairs( ChronoBars_Settings.profiles ) do
    for g=1,table.getn( profile.groups ) do
      local grp = profile.groups[g];
      
      if (grp.sorting == nil) then
        grp.sorting = CB.DEFAULT_GROUP.sorting;
      end

      for b=1,table.getn( profile.groups[g].bars ) do
        local bar = profile.groups[g].bars[b];
        
        if (bar.style.sparkHeight == nil) then
          bar.style.sparkHeight = CB.DEFAULT_BAR.style.sparkHeight;
        end
        
        if (bar.style.sparkWidth == nil) then
          bar.style.sparkWidth = CB.DEFAULT_BAR.style.sparkWidth;
        end
        
        if (bar.style.visibility == nil) then
          bar.style.visibility = CB.DEFAULT_BAR.style.visibility;
        end
        
      end
    end
  end
end

function ChronoBars.UpgradeChar_1_6 ()
end

-- Check for upgrade

function ChronoBars.CheckSettings()
   
  --Check if global settings missing (first time use)
  if (not ChronoBars_Settings) then
    ChronoBars.Print( "Applying default settings" );
    ChronoBars_Settings = CopyTable( ChronoBars.DEFAULT_SETTINGS );
  end
  
  --Check if character settings missing (first time use)
  if (not ChronoBars_CharSettings) then
    ChronoBars.Print( "Applying default character settings" );
    ChronoBars_CharSettings = CopyTable( ChronoBars.DEFAULT_CHAR_SETTINGS );
    
    --Create profile for this character first time if missing
    local playerName = UnitName( "player" );
    if (ChronoBars_Settings.profiles[ playerName ] == nil) then
      ChronoBars_Settings.profiles[ playerName ] = CopyTable( ChronoBars.DEFAULT_PROFILE );
    end
    
    --Switch to profile with this character name
    ChronoBars_CharSettings.activeProfile = playerName;
  end
  
  --Check for old settings version
  local curVersion = ChronoBars_Settings.version;
  for i=1,table.getn( ChronoBars.UPGRADE_LIST ) do
    
    local nextVersion = ChronoBars.UPGRADE_LIST[i];
    if (curVersion < nextVersion) then
      
      ChronoBars.Print( "Upgrading settings from version "
        ..curVersion.." to "..nextVersion );
      
      local verString = string.gsub( nextVersion, "[.]", "_" );
      local funcName = "Upgrade_"..verString;
      local func = ChronoBars[ funcName ];
      
      if (func) then func() else
        ChronoBars.Print( "Missing upgrade function!" );
      end
      
      curVersion = nextVersion;
    end
  end
  
  --Check for old char settings version
  local curVersion = ChronoBars_CharSettings.version;
  for i=1,table.getn( ChronoBars.UPGRADE_LIST ) do
    
    local nextVersion = ChronoBars.UPGRADE_LIST[i];
    if (curVersion < nextVersion) then
      
      ChronoBars.Print( "Upgrading character settings from version "
        ..curVersion.." to "..nextVersion );
      
      local verString = string.gsub( nextVersion, "[.]", "_" );
      local funcName = "UpgradeChar_"..verString;
      local func = ChronoBars[ funcName ];
      
      if (func) then func() else
        ChronoBars.Print( "Missing upgrade function!" );
      end
      
      curVersion = nextVersion;
    end
  end

  --Set to latest version
  ChronoBars_Settings.version = ChronoBars.VERSION;
  ChronoBars_CharSettings.version = ChronoBars.VERSION;
  
  --Switch to default profile if active profile is missing
  if (ChronoBars_Settings.profiles[ ChronoBars_CharSettings.activeProfile ] == nil) then
    ChronoBars_CharSettings.activeProfile = "Default";
  end
  
end


-- LibSharedMedia support
--===================================================

ChronoBars.LSM = LibStub("LibSharedMedia-3.0", true);


--Output functions
--===================================================

function ChronoBars.Print (msg)
  print( "|cffffff00"..ChronoBars.MSG_PREFIX.." |cffffffff"..msg );
end

function ChronoBars.Debug (msg)
  if not ChronoBars_Settings.debugEnabled then return end
  print( "|cffff0000"..ChronoBars.MSG_PREFIX.." |cffffffff"..msg );
end


--Initialization
--===================================================

function ChronoBars.MainFrame_OnEvent (self, event, ...)

  local handlerFnName = "MainFrame_" .. event;
  local handlerFn = ChronoBars[ handlerFnName ];
  if (handlerFn) then handlerFn(...) end
  
end

function ChronoBars.MainFrame_OnUpdate()

  --This gets global game time
  local now = GetTime();

  --First time init
  if (not ChronoBars.initUpdate) then
    ChronoBars.lastUpdate = now;
    ChronoBars.initUpdate = true;
    return
  end

  --Don't update until interval elapsed
  local interval = now - ChronoBars.lastUpdate;
  if (interval < ChronoBars_Settings.updateInterval) then
    return
  end

  --Update now
  ChronoBars.lastUpdate = now;
  ChronoBars.Update( now, interval );
  
  --Clear spellcast list
  for spell,cast in pairs( ChronoBars.spellcastList ) do
    ChronoBars.spellcastList[ spell ] = false;
  end

end


function ChronoBars.Init ()
  
  --The only purpose of main frame is to handle initialization events and global update
  ChronoBars.frame = CreateFrame( "Frame", "ChronoBars_MainFrame" );
  ChronoBars.frame:SetScript( "OnUpdate", ChronoBars.MainFrame_OnUpdate );
  ChronoBars.frame:SetScript( "OnEvent", ChronoBars.MainFrame_OnEvent );
  ChronoBars.frame:RegisterEvent( "ADDON_LOADED" );
  ChronoBars.frame:RegisterEvent( "PLAYER_LOGIN" );
  ChronoBars.frame:RegisterEvent( "DISPLAY_SIZE_CHANGED" );
  ChronoBars.frame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" );
  ChronoBars.frame:RegisterEvent( "ACTIVE_TALENT_GROUP_CHANGED" );

end

--Initialization flags
ChronoBars.initUI = false;
ChronoBars.initUpdate = false;
ChronoBars.lastUpdate = 0;
ChronoBars.designMode = false;
ChronoBars.spellcastList = {};

--Main entry point
ChronoBars.Init ();


--Events
--=====================================================

function ChronoBars.MainFrame_ADDON_LOADED( addonName, ... )

  --Check that it's really us being loaded
  if (addonName ~= "ChronoBars") then return end
  
  --Hmmmmmm doesn't seem to work correctly from here :/
  --ChronoBars.CheckSettings();
end

function ChronoBars.MainFrame_PLAYER_LOGIN (...)
  
  ChronoBars.CheckSettings();
  ChronoBars.UpdateSettings();
  ChronoBars.initUI = true;  
  ChronoBars.Update( GetTime(), 0 );
  
end

function ChronoBars.MainFrame_DISPLAY_SIZE_CHANGED (...)
  if (ChronoBars.initUI) then
    ChronoBars.UpdateSettings();
    ChronoBars.Update( GetTime(), 0 );
  end
end

function ChronoBars.MainFrame_UNIT_SPELLCAST_SUCCEEDED (unit, spell, rank)
  if (unit == "player") then
    ChronoBars.spellcastList[ spell ] = true;
  end
end

function ChronoBars.MainFrame_ACTIVE_TALENT_GROUP_CHANGED ()
  local spec = GetActiveTalentGroup( false, false );

  --Find name of the linked profile
  local specProfile = nil;
  if (spec == 1)
  then specProfile = ChronoBars_CharSettings.primaryProfile;
  else specProfile = ChronoBars_CharSettings.secondaryProfile;
  end

  --Check that profile is linked and exists
  if (specProfile ~= nil) then
    if (ChronoBars_Settings.profiles[ specProfile ] ~= nil) then
    
      --Switch to linked profile
      ChronoBars.Print( "Switching profile to '" .. specProfile .. "' to match spec." );
      ChronoBars_CharSettings.activeProfile = specProfile;
      CB.UpdateSettings();
      CB.Update( GetTime(), 0 );
    end
  end

end


--Slash handler
--===================================================

function ChronoBars.SlashHandler (msg)

  if (msg == "") then
    ChronoBars.designMode = (not ChronoBars.designMode);
    if (ChronoBars.designMode) then
      ChronoBars.Print( "Switching to |cffff0000SETUP |cffffffffmode" );
      ChronoBars.UpdateSettings();
    else
      ChronoBars.Print( "Switching to |cff00ff00RUN |cffffffffmode" );
      ChronoBars.Update( GetTime(), 0 );
    end
    
  elseif (msg == "debug") then
    ChronoBars_Settings.debugEnabled = true;
    ChronoBars.Print( "Debug output |cff00ff00enabled" );
    
  elseif (msg == "nodebug") then
    ChronoBars_Settings.debugEnabled = false;
    ChronoBars.Print( "Debug output |cffff0000disabled" );
         
  else
  
    local cmd, param = strsplit( " ", msg );
    if (cmd and param) then
    
      if (cmd == "update") then
      
        local interval = tonumber(param);
        if (interval and interval > 0) then
          ChronoBars.Print( "Setting update interval to "..tostring(interval) );
          ChronoBars_Settings.updateInterval = interval;
        end
      
      elseif (cmd == "reset") then
      
        if (param == "profile") then
          ChronoBars.ShowConfirmFrame( "Are you sure you want to reset current profile?",
            ChronoBars.ResetProfile_Accept );
            
        elseif (param == "groups") then
          ChronoBars.ShowConfirmFrame( "Are you sure you want to reset group positions?",
            ChronoBars.ResetGroups_Accept );
        end
      end
    end
  end
end

SLASH_ChronoBars1 = "/chronobars";
SLASH_ChronoBars2 = "/cb";
SlashCmdList["ChronoBars"] = ChronoBars.SlashHandler;


--Bar UI
--===================================================

function ChronoBars.ScreenToPixelX (coord)
  return coord * CB.pxRatioX;
end

function ChronoBars.ScreenToPixelY (coord)
  return coord * CB.pxRatioY;
end

function ChronoBars.RoundToPixelX (coord)
  return math.floor( coord * CB.pxRatioX + 0.5 ) / CB.pxRatioX;
end

function ChronoBars.RoundToPixelY (coord)
  return math.floor( coord * CB.pxRatioY + 0.5 ) / CB.pxRatioY;
end

--Assuming pixel ratio will be the same in X and Y since WoW preserves
--aspect ratio between the real window and the scaled UI coordinates
ChronoBars.RoundToPixel = ChronoBars.RoundToPixelX;

function ChronoBars.FindPixelRatio ()

  --Parse resolution X and resolution Y
  local resolutions = {GetScreenResolutions()};
  local res = resolutions[ GetCurrentResolution() ];
  local x = strfind( res, "x" );
  local resX = strsub( res, 1, x-1 );
  local resY = strsub( res, x+1 );
  CB.resX = tonumber( resX );
  CB.resY = tonumber( resY );
  
  --Get the ratio between UI coordinates and actual pixels
  CB.screenW = GetScreenWidth();
  CB.screenH = GetScreenHeight();
  CB.pxRatioX = CB.resX / CB.screenW;
  CB.pxRatioY = CB.resY / CB.screenH;
  
  --Don't use pixel-perfect correction if not full-screen
  CB.windowed = (GetCVar( "gxWindow" ) == "1");
  CB.maximized = (GetCVar( "gxMaximize" ) == "1" );
  if (CB.windowed and (not CB.maximized)) then
    CB.pxRatioX = 1;
    CB.pxRatioY = 1;
  end
  
  CB.Debug( "Screen windowed:"..tostring(CB.windowed).." maximized:"..tostring(CB.maximized) );
  CB.Debug( "Pixel ratioX:"..tostring(CB.pxRatioX).." ratioY:"..tostring(CB.pxRatioY) );
  
end

function ChronoBars.FormatSeconds (seconds, format)

  if (format == ChronoBars.TIME_MINSEC) then
  
    local minutes = 0;
    seconds = math.ceil( seconds );
    
    if (seconds >= 60) then
      minutes = math.floor( seconds / 60 );
      seconds = math.ceil( seconds - minutes * 60 );
    end
    
    return string.format( "%i:%02i", minutes, seconds );
    
  else
  
    local time = seconds;
    local letter = "s";
    local fmt;
        
    if (time >= 60) then
      time = time / 60;
      letter = "m";
    end
    
    if (format == ChronoBars.TIME_DECIMAL) then
      time = math.ceil( time * 10 ) / 10;
      fmt = "%.1f";
    else
      time = math.ceil( time );
      fmt = "%i";
    end
    
    return string.format( fmt..letter, time);
    
  end
end

function ChronoBars.FormatTime (bar, time, fixed)

  local set = bar.settings;
  local timeString = "";
  
  if (set.style.showTime) then
    if (set.type ~= ChronoBars.EFFECT_TYPE_USABLE) then
    
      if (time) then
        timeString = ChronoBars.FormatSeconds( time, set.style.format );
      end

      if (fixed) then
        if (set.fixed.enabled) then
          local fixedString = ChronoBars.FormatSeconds( fixed, set.style.format );
          timeString = timeString.."/"..fixedString;
        end
      end
    end
  end

  return timeString;
end

function ChronoBars.FormatName (bar, name, id, count, order)

  local set = bar.settings;
  local nameString = "";

  if (set.style.showName) then
  
    if (set.display.enabled) then
      nameString = set.display.name;
    elseif (name) then
      nameString = name;
    end

    if (id) then
      nameString = "("..tostring( id )..") "..nameString;
    end
  end

  if (count and count > 1) then
    if (set.type == ChronoBars.EFFECT_TYPE_AURA) then
      if (set.style.showCount) then
        if (set.style.countSide == ChronoBars.SIDE_LEFT)
        then nameString = "["..tostring(count).."]  "..nameString;
        else nameString = nameString.."  ["..tostring(count).."]";
        end
      end
    end
  end

  if (order and order > 1) then
    if (set.type == ChronoBars.EFFECT_TYPE_AURA) then
      nameString = nameString.." #"..tostring(order);
    end
  end

  if (set.type == ChronoBars.EFFECT_TYPE_CD) then
    if (set.style.showCd) then
      nameString = nameString.." CD";
    end
  end
  
  if (set.type == ChronoBars.EFFECT_TYPE_USABLE) then
    if (set.style.showUsable) then
      nameString = nameString.." Usable";
    end
  end

  return nameString;
end


function ChronoBars.GetActiveProfile ()
  return ChronoBars_Settings.profiles[ ChronoBars_CharSettings.activeProfile ];
end


function ChronoBars.Bar_OnMouseDown (bar, button)
  if (button == "RightButton" and ChronoBars.designMode) then
    ChronoBars.OpenBarMenu (bar);
  end
end


function ChronoBars.Bar_OnDragStart (bar)

  if (ChronoBars.designMode) then
    bar.group:StartMoving();
  end
end

function ChronoBars.Bar_OnDragStop (bar)

  if (ChronoBars.designMode) then
    bar.group:StopMovingOrSizing();
    local x = bar.group:GetLeft();
    local y = bar.group:GetBottom();
    bar.group.settings.x = x - GetScreenWidth()/2;
    bar.group.settings.y = y - GetScreenHeight()/2;
    ChronoBars.UpdateSettings();
  end
end

function ChronoBars.UpdateGroupBounds( bar, L,R,B,T )
  
  --We need to move an anchor point a little to refresh bounds
  local p1,f,p2,x,y = bar.bg:GetPoint( 1 );
  bar.bg:SetPoint( p1,f,p2,x+1,y);
  bar.bg:SetPoint( p1,f,p2,x,y);
  
  local l = CB.RoundToPixel( bar.bg:GetLeft() );
  local r = CB.RoundToPixel( bar.bg:GetRight() );
  local t = CB.RoundToPixel( bar.bg:GetTop() );
  local b = CB.RoundToPixel( bar.bg:GetBottom() );
  if (l < L) then L = l end;
  if (b < B) then B = b end;
  if (r > R) then R = r end;
  if (t > T) then T = t end;
  return L,R,B,T;
  
end


function ChronoBars.Group_Create (name, parent, profile, groupId)

  local settings = profile.groups[ groupId ];  
  
  local grp = CreateFrame( "Frame", name, parent );
  grp:SetPoint( "BOTTOMLEFT", parent, "BOTTOMLEFT", settings.x, settings.y );
  grp:SetWidth( 2 );
  grp:SetHeight( 2 );
  grp:Hide();
  
  local bg = grp:CreateTexture( nil, "BACKGROUND" );
  grp.bg = bg;
  
  grp:EnableMouse( true );
  grp:SetMovable( true );
  
  grp.groupId = groupId;
  grp.settings = settings;
  grp.bars = {};
  grp.sortedBars = {};
  return grp;
  
end

function ChronoBars.Group_UpdateSettings (grp, profile, groupId)

  --Get group settings
  local settings = profile.groups[ groupId ];
  
  --Store handy references
  grp.groupId = groupId;
  grp.settings = settings;
  
  --Init complex variables
  local x = CB.RoundToPixel( GetScreenWidth()/2 + settings.x );
  local y = CB.RoundToPixel( GetScreenHeight()/2 + settings.y );
  local w = CB.RoundToPixel( settings.width );
  local h = CB.RoundToPixel( settings.height );
  local s = CB.RoundToPixel( settings.spacing );
  local m = CB.RoundToPixel( settings.margin );
  local p = CB.RoundToPixel( settings.padding );

  --Position group
  grp:ClearAllPoints();
  grp:SetPoint( "BOTTOMLEFT", UIParent, "BOTTOMLEFT", x,y );

  --Position bars
  local L=x; local R=x; local B=y; local T=y;
  local numBars = table.getn( settings.bars );
  
  for b = 1, numBars do
  
    grp.bars[b]:SetWidth( w );
    grp.bars[b]:SetHeight( h );
    
    local offset = (b-1) * (h+s);
    grp.bars[b]:ClearAllPoints();
    grp.bars[b]:SetPoint( "LEFT", 0,0 );
    
    if (settings.grow == ChronoBars.GROW_UP)
    then grp.bars[b]:SetPoint( "BOTTOM", 0, offset );
    else grp.bars[b]:SetPoint( "TOP", 0, -offset );
    end

    L,R,B,T = CB.UpdateGroupBounds( grp.bars[b], L,R,B,T );
  end
  
  --Back color
  local bgcol = settings.style.bgColor;
  grp.bg:SetTexture( bgcol.r, bgcol.g, bgcol.b, bgcol.a );

  --Stretch background to cover all bars
  grp.bg:ClearAllPoints();
  grp.bg:SetPoint( "BOTTOMLEFT", grp, "BOTTOMLEFT", L-x-m, B-y-m );
  grp.bg:SetPoint( "TOPRIGHT",   grp, "BOTTOMLEFT", R-x+m, T-y+m );
  grp.bg:Show();
  
end

function ChronoBars.Bar_Create (name, parent, profile, groupId, barId)

  local gsettings = profile.groups[ groupId ];
  local settings = profile.groups[ groupId ].bars[ barId ];

  local bar = CreateFrame( "Frame", name, parent );

  local bg = bar:CreateTexture( nil, "BACKGROUND" );
  bar.bg = bg;

  local fg = bar:CreateTexture( nil, "ARTWORK" );
  bar.fg = fg;
  
  local fgBlink = bar:CreateTexture( nil, "ARTWORK" );
  bar.fgBlink = fgBlink;
  
  local fontPath = ChronoBars.LSM:Fetch( "font", settings.style.lsmFontHandle );
  if (not fontPath) then
    fontPath = "Fonts\\FRIZQT__.TTF";
  end
  
  local txtName = bar:CreateFontString( name.."txtname" );
  txtName:SetFont( fontPath, settings.style.fontSize );
  txtName:SetJustifyH( "LEFT" );
  txtName:SetTextColor( 1.0, 1.0, 1.0, 1.0 );
  txtName:SetShadowOffset( 1, -1 );
  txtName:SetWordWrap( false );
  bar.txtName = txtName;
  
  local txtTime = bar:CreateFontString (name.."txttime" );
  txtTime:SetFont( fontPath, settings.style.fontSize );
  txtTime:SetJustifyH( "RIGHT" );
  txtTime:SetShadowOffset( 1, -1 );
  txtTime:SetWordWrap( false );
  bar.txtTime = txtTime;
  
  local icon = bar:CreateTexture( nil, "ARTWORK" );
  bar.icon = icon;
  
  local spark = bar:CreateTexture( nil, "OVERLAY" );
  spark:SetTexture( "Interface\\CastingBar\\UI-CastingBar-Spark" );
  spark:SetBlendMode( "ADD" );
  spark:SetWidth( 32 );
  spark:SetHeight( 60 );
  bar.spark = spark;
  
  bar:SetPoint( "LEFT", 0,0 );
  bar:SetPoint( "BOTTOM", 0,0 );
  bar:SetWidth( gsettings.width );
  bar:SetHeight( gsettings.height );
  bar:Hide();
  
  bar.group = parent;
  bar:EnableMouse( true );
  bar:RegisterForDrag( "LeftButton" );
  bar:SetScript( "OnDragStart", ChronoBars.Bar_OnDragStart );
  bar:SetScript( "OnDragStop", ChronoBars.Bar_OnDragStop );
  bar:SetScript( "OnMouseDown", ChronoBars.Bar_OnMouseDown );
  
  bar.groupId = groupId;
  bar.barId = barId;
  bar.gsettings = gsettings;
  bar.settings = settings;
  return bar;
  
end

function ChronoBars.Bar_UpdateSettings (bar, profile, groupId, barId)

  --Get bar and group settings
  local gsettings = profile.groups[ groupId ];
  local settings = profile.groups[ groupId ].bars[ barId ];

  --Store handy references
  bar.groupId = groupId;
  bar.barId = barId;
  bar.gsettings = gsettings;
  bar.settings = settings;
  
  --Reset bar animation
  if (bar.anim == nil) then bar.anim = {} end
  bar.anim.ratio = 0;
  bar.anim.blink = 0;
  bar.anim.fade = 0;
  
  --Reset bar status
  if (bar.status == nil) then bar.status = {} end
  bar.status.icon = nil;
  bar.status.text = "";
  bar.status.duration = nil;
  bar.status.expires = nil;
  bar.status.left = 0.0;
  bar.status.ratio = 0.0;
  bar.status.active = false;
  bar.status.reactive = false;
  bar.status.visible = true;

  --Init effect status for every name in the comma-separated list
  bar.effectStatus = {};
  local effectList = { strsplit( ",", settings.name ) };
  for i,n in ipairs( effectList ) do
  
    local setName = strtrim( n );
    local estatus = {};

    --Check if working with item or spell
    estatus.isItem =
    (
      (settings.type == ChronoBars.EFFECT_TYPE_CD and
      settings.cd.type == ChronoBars.CD_TYPE_ITEM)
      or
      (settings.type == ChronoBars.EFFECT_TYPE_USABLE and
      settings.usable.type == ChronoBars.USABLE_TYPE_ITEM)
    );
    
    --Translate item or spell ID to name and find its icon
    estatus.id = tonumber( setName );
    if (estatus.id) then

      --These don't work for spells in player's spellbook if name given
      if (estatus.isItem) then
        estatus.name, _, _, _, _, _, _, _, _,
        estatus.icon = GetItemInfo( setName );
      else
        estatus.name, _,
        estatus.icon = GetSpellInfo( setName );
      end
    else

      --These work for spells in player's spellbook with name but not with id
      estatus.name = setName;
      if (estatus.isItem)
      then estatus.icon = GetItemIcon( setName );
      else estatus.icon = GetSpellTexture( setName );
      end
    end
    
    --Reset effect status
    estatus.text = nil;
    estatus.duration = nil;
    estatus.expires = nil;
    estatus.usableCdExpires = nil;
    
    --Add to status list
    table.insert( bar.effectStatus, estatus );
    
    --Show icon of the first spell
    if (i == 1) then
      bar.status.icon = estatus.icon;
    end
  end
  
  --Round graphics to actual pixels
  local pad = CB.RoundToPixel( gsettings.padding );
  local h = CB.RoundToPixel( gsettings.height );
  local w = CB.RoundToPixel( gsettings.width );
  
  --Fetch LibSharedMedia paths
  local fontPath = ChronoBars.LSM:Fetch( "font", settings.style.lsmFontHandle );
  if (not fontPath) then
    settings.style.lsmFontHandle = "Friz Quadrata TT";
    fontPath = "Fonts\\FRIZQT__.TTF";
  end

  local texPath = ChronoBars.LSM:Fetch( "statusbar", settings.style.lsmTexHandle );
  if (not texPath) then
    settings.style.lsmTexHandle = "None";
    texPath = "";
  end
  
  --Back and front color and texture
  local b = settings.style.bgColor;
  bar.bg:SetTexture( b.r, b.g, b.b, b.a );
    
  local f = settings.style.fgColor;
  bar.fg:SetHorizTile( true );
  bar.fg:SetVertTile( false );
  bar.fg:SetGradientAlpha( "HORIZONTAL",  f.r,f.g,f.b,f.a,  f.r,f.g,f.b,f.a );
    
  bar.fgBlink:SetHorizTile( true );
  bar.fgBlink:SetVertTile( false );
  bar.fgBlink:SetGradientAlpha( "HORIZONTAL", f.r,f.g,f.b,0,  f.r,f.g,f.b,0 );
    
  if (settings.style.lsmTexHandle == "None") then
  
    bar.fg:SetTexture( 1,1,1,1 );
    bar.fgBlink:SetTexture( 1,1,1,1 );
    
  else
    bar.fg:SetTexture( texPath );
    bar.fgBlink:SetTexture( texPath );
    
  end
  
  --Make foreground half-full
  local offWidth = (w - 2*pad) * 0.5;
  bar.fg:ClearAllPoints();
  bar.fgBlink:ClearAllPoints();

  if (settings.style.fullSide == CB.SIDE_RIGHT) then
    bar.fg:SetPoint( "BOTTOMLEFT", pad, pad );
    bar.fg:SetPoint( "TOPRIGHT", -pad-offWidth, -pad );
    bar.fgBlink:SetPoint( "BOTTOMLEFT", bar.fg, "BOTTOMRIGHT", 0,0 );
    bar.fgBlink:SetPoint( "TOPRIGHT", -pad, -pad );
  else
    bar.fg:SetPoint( "BOTTOMLEFT", pad+offWidth, pad );
    bar.fg:SetPoint( "TOPRIGHT", -pad, -pad );
    bar.fgBlink:SetPoint( "BOTTOMLEFT", pad, pad );
    bar.fgBlink:SetPoint( "TOPRIGHT", bar.fg, "TOPLEFT", 0,0 );
  end
  
  --Time text
  local timeString;
  if (settings.type == ChronoBars.EFFECT_TYPE_CUSTOM)
  then timeString = ChronoBars.FormatTime( bar, settings.custom.duration, settings.fixed.duration );
  else timeString = ChronoBars.FormatTime( bar, 0, settings.fixed.duration );
  end
  
  bar.txtTime:SetText( timeString );

  local tcol = settings.style.textColor;
  bar.txtTime:SetTextColor( tcol.r, tcol.g, tcol.b, tcol.a );
  bar.txtTime:SetFont( fontPath, settings.style.fontSize );
  bar.txtTime:ClearAllPoints();

  if (settings.style.timeSide == CB.SIDE_RIGHT) then
    bar.txtTime:SetPoint( "BOTTOMRIGHT", -pad - 2, pad );
    bar.txtTime:SetPoint( "TOPRIGHT", -pad - 2, -pad );
    bar.txtTime:SetJustifyH( "RIGHT" );

  elseif (settings.style.timeSide == CB.SIDE_LEFT) then
    bar.txtTime:SetPoint( "BOTTOMLEFT", pad + 2, pad );
    bar.txtTime:SetPoint( "TOPLEFT", pad + 2, -pad );
    bar.txtTime:SetJustifyH( "LEFT" );
  end
  
  --Name / id / count text
  local nameString = ChronoBars.FormatName( bar, settings.name, nil, 2, settings.aura.order );
  bar.txtName:SetText( nameString );

  local tcol = settings.style.textColor;
  bar.txtName:SetTextColor( tcol.r, tcol.g, tcol.b, tcol.a );
  bar.txtName:SetFont( fontPath, settings.style.fontSize );
  bar.txtName:ClearAllPoints();
  
  if (settings.style.timeSide == CB.SIDE_RIGHT)
  then bar.txtName:SetJustifyH( "LEFT" );
  else bar.txtName:SetJustifyH( "RIGHT" );
  end
  
  if (settings.style.timeSide == CB.SIDE_RIGHT) then
    bar.txtName:SetPoint( "BOTTOMLEFT", pad + 2, pad );
    bar.txtName:SetPoint( "TOPRIGHT", bar.txtTime, "TOPLEFT", -2, 0 );
  
  elseif (settings.style.timeSide == CB.SIDE_LEFT) then
    bar.txtName:SetPoint( "BOTTOMLEFT", bar.txtTime, "BOTTOMRIGHT", 2, 0 );
    bar.txtName:SetPoint( "TOPRIGHT", -pad - 2, -pad );
  end

  bar.txtName:Show();
  
  --Icon
  if (settings.style.showIcon) then
    
    --Use ? texture if not found
    local iconTex = bar.status.icon;
    if (not iconTex) then iconTex = "Interface/Icons/INV_Misc_QuestionMark"; end
    bar.icon:SetTexture( iconTex );
    bar.icon:Show();
    
    --Icon zoom
    if (settings.style.iconZoom) then
      local s = 0.08;
      bar.icon:SetTexCoord( s, 1-s, s, 1-s );
    else
      bar.icon:SetTexCoord( 0,1,0,1 );
    end
    
    if (settings.style.iconSide == CB.SIDE_LEFT) then
    
      --Enlarge background to cover the icon
      bar.bg:SetPoint( "TOPRIGHT", 0,0 );
      bar.bg:SetPoint( "BOTTOMLEFT", -h + pad, 0);
      
      --Set icon to the left of the bar
      bar.icon:SetPoint( "TOPRIGHT", bar, "TOPLEFT", 0, -pad );
      bar.icon:SetPoint( "BOTTOMLEFT", bar, "BOTTOMLEFT", -h + 2*pad, pad );
    
    elseif (settings.style.iconSide == CB.SIDE_RIGHT) then
    
      --Enlarge background to cover the icon
      bar.bg:SetPoint( "TOPRIGHT", h - pad,0 );
      bar.bg:SetPoint( "BOTTOMLEFT", 0, 0);
      
      --Set icon to the left of the bar
      bar.icon:SetPoint( "TOPRIGHT", bar, "TOPRIGHT", h - 2*pad, -pad );
      bar.icon:SetPoint( "BOTTOMLEFT", bar, "BOTTOMRIGHT", 0, pad );
      
    end
    
  else
    --Shrink background to bar only
    bar.bg:SetPoint( "TOPRIGHT", 0,0 );
    bar.bg:SetPoint( "BOTTOMLEFT", 0,0 );
    bar.icon:Hide();
  end
  
  --Spark
  if (settings.style.showSpark) then
    bar.spark:SetWidth( settings.style.sparkWidth );
    bar.spark:SetHeight( h * settings.style.sparkHeight );
    bar.spark:Show();
    
    if (settings.style.fullSide == CB.SIDE_RIGHT)
    then bar.spark:SetPoint( "CENTER", bar.fg, "RIGHT", 0,0 );
    else bar.spark:SetPoint( "CENTER", bar.fg, "LEFT", 0,0 );
    end
  else
    bar.spark:Hide();
  end
  
  --Reset fading
  bar:SetAlpha( 1 );
  
  --Enable for mouse events
  bar:EnableMouse( true );
  
end

function ChronoBars.UpdateSettings ()

  --Update pixel-perfect state
  CB.FindPixelRatio();
  
  --Get the character's active profile
  local profile = ChronoBars.GetActiveProfile();
  
  --Create group table if missing
  if (not ChronoBars.groups) then
    ChronoBars.groups = {};
  end

  --Walk profile groups
  for g = 1, table.getn( profile.groups ) do
  
    --Create group frame if missing
    if (not ChronoBars.groups[g]) then
      ChronoBars.groups[g] = ChronoBars.Group_Create(
        "ChronoBars_Group"..tostring(g),
        UIParent, profile, g
      );
    end
    
    --Walk profile bars
    for b = 1, table.getn( profile.groups[g].bars ) do
      
      --Create bar frame if missing
      if (not ChronoBars.groups[g].bars[b]) then
        ChronoBars.groups[g].bars[b] = ChronoBars.Bar_Create(
          "ChronoBars_Group"..tostring(g).."Bar"..tostring(b),
          ChronoBars.groups[g], profile, g, b
        );
      end
      
      --Update bar frame
      ChronoBars.groups[g].bars[b]:Show();
      ChronoBars.Bar_UpdateSettings( ChronoBars.groups[g].bars[b], profile, g, b );
      
    end
    
    --Update group frame (after bars to position them)
    ChronoBars.groups[g]:Show();
    ChronoBars.Group_UpdateSettings( ChronoBars.groups[g], profile, g );
    
    --Hide unused bars
    for b = table.getn( profile.groups[g].bars )+1, table.getn( ChronoBars.groups[g].bars ) do
      ChronoBars.groups[g].bars[b]:Hide();
    end
    
  end
  
  --Hide unused groups
  for g = table.getn( profile.groups )+1, table.getn( ChronoBars.groups ) do
    ChronoBars.groups[g]:Hide();
  end

end

--Aura
--=================================================

function ChronoBars.Bar_UpdateTimeAura (bar, status, now)

  local set = bar.settings;
  local name, icon, count, duration, expires, caster, id;
  count = 0;

  --Setup buff/debuff filter
  local filter = set.aura.type;
  if (set.aura.byPlayer) then
    filter = filter.."|PLAYER";
  end

  --UnitAura doesn't take spell ID :(
  --If using ID or summing stacks we need to check all auras
  if (set.aura.sum) then

    --Walk all the auras on the unit matching the filter
    local auraIndex = 1;
    while true do

      --Get aura info by index
      local tempCount, tempDur, tempExp;
      local tempName, _, tempIcon, tempCount, _, tempDur, tempExp, tempCaster, _, _, tempId =
        UnitAura( set.aura.unit, auraIndex, filter );
      if (not tempName) then break end;
      auraIndex = auraIndex + 1;

      --Check if ID or name matches
      local match;
      if (status.id)
      then match = (tempId == status.id);
      else match = (tempName == status.name);
      end

      --Add to count, pick largest duration and expiration
      if (match) then
        name = tempName; icon = tempIcon; id = tempId;

        if (not tempCount) then tempCount = 1; end;
        count = count + tempCount;

        if ((not duration) or tempDur > duration) then
          duration = tempDur;
        end

        if ((not expires) or tempExp > expires) then
          expires = tempExp;
        end
      end
    end

  elseif (status.id or set.aura.order > 1) then

    --Walk all the auras on the unit matching the filter
    local auraOrder = 1;
    local auraIndex = 1;
    while true do

      --Get aura info by index
      name, _, icon, count, _, duration, expires, caster, _, _, id =
        UnitAura( set.aura.unit, auraIndex, filter );
      if (not name) then break end;
      auraIndex = auraIndex + 1;

      --Check if ID or name matches
      local match;
      if (status.id)
      then match = (id == status.id);
      else match = (name == status.name);
      end

      --Check if ID matches
      if (match) then
        if (auraOrder == set.aura.order) then break end;
        auraOrder = auraOrder + 1;
      end
    end

  else
    --Get aura info by name
    name, _, icon, count, _, duration, expires =
      UnitAura( set.aura.unit, status.name, nil, filter );
  end
  
  --Update bar time
  status.duration = duration;
  status.expires = expires;

  --Update bar icon
  if (icon) then
    status.icon = icon;
  end
  
  --Update bar text with stack count
  status.text = ChronoBars.FormatName( bar, status.name, nil, count );
  
end

--Cooldown
--===================================================
  
function ChronoBars.Bar_UpdateTimeCd (bar, status, now)

  local set = bar.settings;
  local start = nil;
  local duration = nil;

  --Get cooldown times (works with name and ID)
  if (status.isItem)
  then start, duration = GetItemCooldown( status.name );
  else start, duration = GetSpellCooldown( status.name );
  end

  --Update bar time unless it is a global cooldown
  if (start == nil or duration == nil) then
    status.expires = nil;
    status.duration = nil;
    
  elseif( start == 0 or duration == 0) then
    status.expires = nil;
    status.duration = nil;
        
  elseif (duration > 1.5) then
    status.expires = start + duration;
    status.duration = duration;
  end

  --Update bar text
  status.text = ChronoBars.FormatName( bar, status.name, nil, nil );
  
end

--Usable
--===================================================

function ChronoBars.Bar_UpdateTimeUsable (bar, status, now)

  local set = bar.settings;
  local usable, start, duration;

  --Check if it is usable in general
  if (status.isItem)
  then usable = IsUsableItem( status.name );
  else usable = IsUsableSpell( status.name );
  end

  if (set.usable.includeCd) then
  
    --Get spell or item cooldown
    if (status.isItem)
    then start, duration = GetItemCooldown( status.name );
    else start, duration = GetSpellCooldown( status.name );
    end
    
    --Update usable cd unless it is global cooldown
    if (start == nil or duration == nil) then
      status.usableCdExpires = nil;
    elseif (start == 0 or duration == 0) then
      status.usableCdExpires = nil;
    elseif (duration > 1.5) then
      status.usableCdExpires = start + duration;
    end
    
    --Check if usable cd ran out
    if (status.usableCdExpires) then
      if (now > status.usableCdExpires) then
        status.usableCdExpires = nil;
      end
    end
    
    --Include in usability check
    usable = (usable and (status.usableCdExpires==nil));
  end
  
  --Update bar time
  if (usable) then
    status.duration = 0;
    status.expires = 0;
  else
    status.duration = nil;
    status.expires = nil;
  end
  
  --Update bar text
  status.text = ChronoBars.FormatName( bar, status.name, nil, nil );

end


--Totem
--================================================
  
function ChronoBars.Bar_UpdateTimeTotem (bar, status, now)

  local set = bar.settings;
  local _, name, start, duration, icon = GetTotemInfo( set.totem.type );

  --Update bar time
  if (name and start and duration) then
    --The name we get includes totem rank so search for substring
    if (strfind( name , status.name )) then
      status.expires = start + duration;
      status.duration = duration;
    else
      status.expires = nil;
      status.duration = nil;
    end
  else
    status.expires = nil;
    status.duration = nil;
  end

  --Update bar icon
  if (icon) then
    status.icon = icon;
  end
  
  --Update bar text
  status.text = ChronoBars.FormatName( bar, status.name, nil, nil );
  
end

--Custom
--===================================================

function ChronoBars.Bar_UpdateTimeCustom (bar, status, now)

  local set = bar.settings;
  local profile = CB.GetActiveProfile();
  local reactive = false;

  if (set.custom.trigger == CB.CUSTOM_TRIGGER_SPELL_CAST) then
  
    --Check if spell was cast since the last update
    if (CB.spellcastList[ status.name ] == true) then
      reactive = true;
    end
    
  elseif (set.custom.trigger == CB.CUSTOM_TRIGGER_BAR_ACTIVE) then
  
    --Walk all the bars in current profile
    for g=1,table.getn( profile.groups ) do
      for b=1,table.getn( profile.groups[g].bars ) do
      
        --Check if other bar is tracking this effect
        local otherSet = profile.groups[g].bars[b];
        if (otherSet.type ~= CB.EFFECT_TYPE_CUSTOM and 
            strfind( otherSet.name, status.name )) then
          
          --Check if bar just got active
          local otherBar = CB.groups[g].bars[b];
          if (otherBar.status.reactive) then
            status.icon = otherBar.icon:GetTexture();
            reactive = true;
          end
        end
      end
    end
  end
  
  --Update bar time if just got active
  if (reactive) then
    status.duration = set.custom.duration;
    status.expires = now + set.custom.duration;
  end
  
  --Update bar text
  status.text = ChronoBars.FormatName( bar, status.name, nil, nil );
  
end


function ChronoBars.Bar_UpdateTime (bar, now, interval)
  
  local p = CB.RoundToPixel( bar.gsettings.padding );
  local w = CB.RoundToPixel( bar.gsettings.width );
  local set = bar.settings;
  
  --Status
  --===============================
  bar.status.duration = nil;
  bar.status.expires = nil;
  
  --Walk the list of effects
  for i=1,table.getn( bar.effectStatus ) do
    local estatus = bar.effectStatus[i];

    --Update duration and expiration times
    if (set.type == CB.EFFECT_TYPE_AURA) then
      CB.Bar_UpdateTimeAura( bar, estatus, now );

    elseif (set.type == CB.EFFECT_TYPE_CD) then
      CB.Bar_UpdateTimeCd( bar, estatus, now );

    elseif (set.type == CB.EFFECT_TYPE_USABLE) then
      CB.Bar_UpdateTimeUsable( bar, estatus, now );

    elseif (set.type == CB.EFFECT_TYPE_TOTEM) then
      CB.Bar_UpdateTimeTotem( bar, estatus, now );

    elseif (set.type == CB.EFFECT_TYPE_CUSTOM) then
      CB.Bar_UpdateTimeCustom( bar, estatus, now );
    end
    
    --Check if this effect is the last to expire
    if (estatus.duration ~= nil and estatus.expires ~= nil) then
      if (bar.status.expires == nil or estatus.expires > bar.status.expires) then
      
        --Copy to bar status
        bar.status.icon = estatus.icon;
        bar.status.text = estatus.text;
        bar.status.duration = estatus.duration;
        bar.status.expires = estatus.expires;
      end
    end
  end
  
  --Update time left and ratio
  bar.status.left = 0.0;
  bar.status.ratio = 0.0;
  
  if (bar.status.duration ~= nil and bar.status.expires ~= nil) then
    if (bar.status.duration == 0 and bar.status.expires == 0) then

      --It's a no-duration bar
      bar.status.ratio = 1.0;

    elseif (now > bar.status.expires) then
    
      --It ran out
      bar.status.duration = nil;
      bar.status.expires = nil;
      
    else
    
      --Check if fixed duration is enabled
      local effectiveDuration = bar.status.duration;
      if (set.fixed.enabled) then
        effectiveDuration = set.fixed.duration;
      end
      
      --Calculate time ratio
      bar.status.left = bar.status.expires - now;
      if (effectiveDuration > 0.0) then
        bar.status.ratio = bar.status.left / effectiveDuration;
      end
      
    end
  end
  
   --Clamp time ratio
  if (bar.status.ratio > 1.0) then
    bar.status.ratio = 1.0;
  elseif (bar.status.ratio < 0.0) then
    bar.status.ratio = 0.0;
  end
  
  
  --Animation
  --===============================
  
  --Animate ratio
  local dRatio = interval * 3.0;
  if (bar.status.ratio > bar.anim.ratio) then

    --Slide up when refreshed
    if (set.style.anim.up) then
      bar.anim.ratio = bar.anim.ratio + dRatio;
      if (bar.anim.ratio > bar.status.ratio) then
        bar.anim.ratio = bar.status.ratio;
      end
    else bar.anim.ratio = bar.status.ratio end
    
  elseif (bar.status.ratio < bar.anim.ratio) then
  
    --Slide down when consumed
    if (set.style.anim.down) then
      bar.anim.ratio = bar.anim.ratio - dRatio;
      if (bar.anim.ratio < bar.status.ratio) then
        bar.anim.ratio = bar.status.ratio;
      end
    else bar.anim.ratio = bar.status.ratio end
  end

  --Find if should blink
  local canBlink = false;
  if (bar.status.duration == 0) then
    canBlink = set.style.anim.blinkUsable;
  else
  
    --Find start time of fading (half bar or 5 seconds if bar longer)
    local blinkStart = 0;
    if (bar.status.duration) then
      blinkStart = 0.5 * bar.status.duration
      if (blinkStart > 5) then blinkStart = 5 end;
    end
  
    --Blink when bar time below threshold  
    canBlink = set.style.anim.blink and (bar.status.left <= blinkStart);
  end
  
  --Animate blinking
  if (canBlink) then
  
    --Alpha is absolute of [1,-1] for back-and-forth animation
    local dBlink = interval * 3.0;
    bar.anim.blink = bar.anim.blink - dBlink;
    if (bar.anim.blink < -1.0) then
      bar.anim.blink = 1;
    end
  else
    bar.anim.blink = 0;
  end
  
  --Animate fading (fade out when bar expires)
  if (set.style.anim.fade) then
    if (bar.status.ratio == 0 and set.style.visibility ~= CB.VISIBLE_ALWAYS) then
      
      local dFade = interval * 3.0;
      bar.anim.fade = bar.anim.fade - dFade;
      if (bar.anim.fade < 0.0) then
        bar.anim.fade = 0;
      end
      
    else bar.anim.fade = 1; end
  else bar.anim.fade = 1; end

  --Bar is active if time is left
  local newActive = (bar.status.ratio > 0.0);
  bar.status.reactive = (newActive and (not bar.status.active));
  bar.status.active = newActive;
    
  --Bar is visible if time is left and animations are done
  bar.status.visible = true;
  if (set.style.visibility ~= CB.VISIBLE_ALWAYS) then
    if (set.style.anim.fade) then
      bar.status.visible = (bar.anim.fade > 0.0);
    elseif (set.style.anim.down) then
      bar.status.visible = (bar.anim.ratio > 0.0);
    else
      bar.status.visible = (bar.status.ratio > 0.0);
    end
  end
  
  --UI
  --===============================
  
  --Hide spark for no-duration and expired bars
  if (bar.status.duration == 0 or bar.anim.ratio == 0) then
    bar.spark:Hide();
  else
    if (set.style.showSpark) then
      bar.spark:Show();
    end
  end
  
  --Scale front to match ratio
  local maxW = w - 2 * p;
  
  local offW;
  if (set.style.fillUp)
  then offW = bar.anim.ratio * maxW;
  else offW = (1.0 - bar.anim.ratio) * maxW;
  end

  if (set.style.fullSide == ChronoBars.SIDE_RIGHT) then
    bar.fg:SetPoint( "TOPRIGHT", -p - offW, -p );
  elseif (set.style.fullSide == ChronoBars.SIDE_LEFT) then
    bar.fg:SetPoint( "BOTTOMLEFT", p + offW, p );
  end
  
  --Set front alpha to match blink
  local blinkA = math.abs( bar.anim.blink );
  local f = set.style.fgColor;
  local fa = 0.5 * blinkA * f.a;

  --Apply blink to full portion if no-duration, exhausted portion otherwise
  if (bar.status.duration == 0) then
    bar.fg:SetGradientAlpha( "HORIZONTAL",
      f.r,f.g,f.b, fa,
      f.r,f.g,f.b, fa );
  else
    bar.fgBlink:SetGradientAlpha( "HORIZONTAL",
      f.r,f.g,f.b, fa,
      f.r,f.g,f.b, fa );
  end
  
  --Set entire bar alpha to match fade
  bar:SetAlpha( bar.anim.fade );
  
  --Update bar icon
  if (bar.status.icon) then
    bar.icon:SetTexture( bar.status.icon );
  end
  
  --Set name text
  bar.txtName:SetText( bar.status.text );
  
  --Set time text
  bar.txtTime:SetText( ChronoBars.FormatTime( bar, bar.status.left ));
  
  --Disable mouse events
  bar:EnableMouse( false );
  
end

function ChronoBars_BarSmaller (a,b)
  if (a.status.left ~= b.status.left) then
    return a.status.left < b.status.left;
  else
    return a.barId < b.barId;
  end
end

function ChronoBars_BarGreater (a,b)
  if (a.status.left ~= b.status.left) then
    return a.status.left > b.status.left;
  else
    return a.barId > b.barId;
  end
end

function ChronoBars.Group_UpdateTime (grp, now, interval)

  local x = CB.RoundToPixel( grp:GetLeft() );
  local y = CB.RoundToPixel( grp:GetBottom() );
  local m = CB.RoundToPixel( grp.settings.margin );
  local h = CB.RoundToPixel( grp.settings.height );
  local s = CB.RoundToPixel( grp.settings.spacing );
  local numBars = table.getn( grp.settings.bars );

  local L=x; local R=x; local B=y; local T=y;
  local numActive = 0;
  
  --Copy bars to sorted array
  for b = 1, numBars do
    grp.sortedBars[ b ] = grp.bars[ b ];
  end
  grp.sortedBars[ numBars+1 ] = nil;
  
  --Sort bars by time left
  if (grp.settings.sorting == CB.SORT_ASCENDING) then
    table.sort( grp.sortedBars, ChronoBars_BarSmaller );
  elseif (grp.settings.sorting == CB.SORT_DESCENDING) then
    table.sort( grp.sortedBars, ChronoBars_BarGreater );
  end
  
  --Update bar positions
  for b = 1, numBars do
    if (grp.sortedBars[b].status.visible) then
    
      --Hide everything but first visible if priority group
      if (grp.settings.layout == CB.LAYOUT_PRIORITY and numActive > 0) then
        grp.sortedBars[b]:Hide();
      else
        
        --Find bar offset
        local offset = 0;
        if (grp.settings.layout == CB.LAYOUT_KEEP) then
          offset = (b-1) * (h+s);
        elseif (grp.settings.layout == CB.LAYOUT_STACK) then
          offset = numActive * (h+s);
        end
        
        --Grow group up or down
        grp.sortedBars[b]:ClearAllPoints();
        grp.sortedBars[b]:SetPoint( "LEFT", 0,0 );
        
        if (grp.settings.grow == ChronoBars.GROW_UP) then
          grp.sortedBars[b]:SetPoint( "BOTTOM", 0, offset );
        else
          grp.sortedBars[b]:SetPoint( "TOP", 0, -offset );
        end
        
        --Show visible bar
        grp.sortedBars[b]:Show();
        numActive = numActive + 1;
        L,R,B,T = CB.UpdateGroupBounds( grp.sortedBars[b], L,R,B,T );
      end
    else
    
      --Hide invisible bar
      grp.sortedBars[b]:Hide();
      
    end
  end
  
  --Stretch background to cover all bars
  if (numActive > 0) then
    grp.bg:SetPoint( "BOTTOMLEFT", grp, "BOTTOMLEFT", L-x-m, B-y-m );
    grp.bg:SetPoint( "TOPRIGHT",   grp, "BOTTOMLEFT", R-x+m, T-y+m );
    grp.bg:Show();
  else
    grp.bg:Hide();
  end
  
end

function ChronoBars.UpdateTime (now, interval)
  
  --Get the character's active profile
  local profile = ChronoBars.GetActiveProfile();
  
  --Walk groups
  local numGroups = table.getn( profile.groups );
  for g = 1, numGroups  do
  
    --Walk non-custom bars
    local numBars = table.getn( profile.groups[g].bars );
    for b = 1, numBars do
    
      --Update will check if its active
      local bar = CB.groups[g].bars[b];
      if (bar.settings.type ~= CB.EFFECT_TYPE_CUSTOM) then
        ChronoBars.Bar_UpdateTime( bar, now, interval );
      end
    end
    
    --Walk custom bars
    for b = 1, numBars do
    
      --Activation might depend on another bar
      local bar = CB.groups[g].bars[b];
      if (bar.settings.type == CB.EFFECT_TYPE_CUSTOM) then
        ChronoBars.Bar_UpdateTime( bar, now, interval );
      end
    end
    
    --Update group (after bars to sort active bars)
    ChronoBars.Group_UpdateTime( ChronoBars.groups[g], now, interval );
  end
  
end

function ChronoBars.Update (now, interval)
  
  if (ChronoBars.initUI) then  
    if (not ChronoBars.designMode) then
      ChronoBars.UpdateTime( now, interval );
    end
  end
  
end
